module REPL where

import Eval
import FileEval
import Lexer
import Parser
import Research

import Control.Exception         (SomeException, catch)
import Control.Monad.IO.Class    (liftIO)
import Control.Monad.Catch       (handle, MonadCatch)
import Control.Monad.Trans.Class (lift)
import Control.Monad.Trans.Maybe (MaybeT(..), runMaybeT)
import Data.Char                 (isSpace)
import Data.List                 ( dropWhile
                                 , dropWhileEnd
                                 , isPrefixOf)
import System.Console.Haskeline

import qualified Data.Map as Map

repl :: Env -> IO ()
repl env = runInputT settings (withInterrupt (loop env Decode))
  where
    settings :: Settings IO
    settings = Settings
      { complete = completeWord Nothing " \t" completeCommands
      , historyFile = Just ".tricu_history"
      , autoAddHistory = True
      }

    completeCommands :: String -> IO [Completion]
    completeCommands str = return $ map simpleCompletion $
      filter (str `isPrefixOf`) commands
      where
        commands = ["!exit", "!output", "!definitions", "!import"]

    loop :: Env -> EvaluatedForm -> InputT IO ()
    loop env form = handle (interruptHandler env form) $ do
      minput <- getInputLine "tricu < "
      case minput of
        Nothing -> outputStrLn "Exiting tricu"
        Just s
          | strip s == "" -> loop env form
          | strip s == "!exit" -> outputStrLn "Exiting tricu"
          | strip s == "!output" -> handleOutput env form
          | strip s == "!definitions" -> do
              let defs = Map.keys $ Map.delete "!result" env
              if null defs
                then outputStrLn "No definitions discovered."
                else do
                  outputStrLn "Available definitions:"
                  mapM_ outputStrLn defs
              loop env form
          | "!import" `isPrefixOf` strip s -> handleImport env form
          | take 2 s == "--" -> loop env form
          | otherwise -> do
              newEnv <- liftIO $ processInput env s form `catch` errorHandler env
              loop newEnv form

    handleOutput :: Env -> EvaluatedForm -> InputT IO ()
    handleOutput env currentForm = do
      let formats = [Decode, TreeCalculus, FSL, AST, Ternary, Ascii]
      outputStrLn "Available output formats:"
      mapM_ (\(i, f) -> outputStrLn $ show i ++ ". " ++ show f)
        (zip [1..] formats)

      result <- runMaybeT $ do
        input <- MaybeT $ getInputLine "Select output format (1-6) < "
        case reads input of
          [(n, "")] | n >= 1 && n <= 6 ->
            return $ formats !! (n-1)
          _ -> MaybeT $ return Nothing

      case result of
        Nothing -> do
          outputStrLn "Invalid selection. Keeping current output format."
          loop env currentForm
        Just newForm -> do
          outputStrLn $ "Output format changed to: " ++ show newForm
          loop env newForm

    handleImport :: Env -> EvaluatedForm -> InputT IO ()
    handleImport env form = do
      result <- runMaybeT $ do
        let fileSettings = setComplete completeFilename defaultSettings
        path <- MaybeT $ runInputT fileSettings $
          getInputLineWithInitial "File path to load < " ("", "")

        contents <- liftIO $ readFile (strip path)

        if | Left err <- parseProgram (lexTricu contents) -> do
               lift $ outputStrLn $ "Parse error: " ++ handleParseError err
               MaybeT $ return Nothing
           | Right ast <- parseProgram (lexTricu contents) -> do
               ns <- MaybeT $ runInputT defaultSettings $
                 getInputLineWithInitial "Namespace (or !Local for no namespace) < " ("", "")

               processedAst <- liftIO $ preprocessFile (strip path)
               let namespacedAst | strip ns == "!Local" = processedAst
                                | otherwise = nsDefinitions (strip ns) processedAst
                   loadedEnv = evalTricu env namespacedAst
               return loadedEnv

      if | Nothing <- result -> do
             outputStrLn "Import cancelled."
             loop env form
         | Just loadedEnv <- result ->
             loop (Map.delete "!result" loadedEnv) form

    interruptHandler :: Env -> EvaluatedForm -> Interrupt -> InputT IO ()
    interruptHandler env form _ = do
      outputStrLn "Interrupted with CTRL+C\n\
                  \You can use the !exit command or CTRL+D to exit"
      loop env form

    processInput :: Env -> String -> EvaluatedForm -> IO Env
    processInput env input form = do
      let asts   = parseTricu input
          newEnv = evalTricu env asts
      case Map.lookup "!result" newEnv of
        Just r -> do
          putStrLn $ "tricu > " ++ formatResult form r
        Nothing -> pure ()
      return newEnv

    errorHandler :: Env -> SomeException -> IO (Env)
    errorHandler env e = do
      putStrLn $ "Error: " ++ show e
      return env

    strip :: String -> String
    strip = dropWhileEnd isSpace . dropWhile isSpace
